// 1次元FFT(16点)のホストコード
// サンプリング周波数2.56x10KHz=25.6KHzとし,
// 8000Hzと1600Hzの合成波をテスト入力として
// FFTを実装したカーネルに与えて、FFTの結果(スペクトル)を
// Excel用に .csvとして出力します.
// ホストはAlteraSoC上のARMプロセッサで実行されるので,
// 出力した.csvファイルをscp等でPC上にコピーして
// Excelで読み込んで表示してください.

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include <cstring>
#include "CL/opencl.h"
#include "AOCLUtils/aocl_utils.h"
#include "fft_config.h"

// FFTの分解能を設定. デフォルトはLOGN=4なので, 2^4=16点
#define N (1 << LOGN)

using namespace aocl_utils;

#define STRING_BUFFER_LEN 1024

// ACL runtime configuration
static cl_platform_id platform = NULL;
static cl_device_id device = NULL;
static cl_context context = NULL;
static cl_command_queue queue = NULL;
static cl_kernel kernel = NULL;
static cl_program program = NULL;
static cl_int status = 0;

// FFT operates with complex numbers - store them in a struct
typedef struct {
  double x;
  double y;
} double2;

typedef struct {
  float x;
  float y;
} float2;

// Function prototypes
bool init();
void cleanup();
static void test_fft(int iterations, bool inverse);
static int coord(int iteration, int i);
static void fourier_transform_gold(bool inverse, int lognr_points, double2 * data);
static void fourier_stage(int lognr_points, double2 * data);


// Host memory buffers
float2 *h_inData, *h_outData;
double2 *h_verify;

// Device memory buffers
cl_mem d_inData, d_outData;

// Entry point.
int main(int argc, char **argv) {
  if(!init()) {
    return false;
  }
  int iterations = 1;

  Options options(argc, argv);

  // Optional argument to set the number of iterations.
  if(options.has("n")) {
    iterations = options.get<int>("n");
  }

  if (iterations <= 0) {
	printf("ERROR: Invalid number of iterations\n\nUsage: %s [-N=<#>]\n\tN: number of iterations to run (default 2000)\n", argv[0]);
	return false;
  }

  // Allocate host memory
  h_inData = (float2 *)alignedMalloc(sizeof(float2) * N * iterations);
  h_outData = (float2 *)alignedMalloc(sizeof(float2) * N * iterations);
  h_verify = (double2 *)alignedMalloc(sizeof(double2) * N * iterations);
  if (!(h_inData && h_outData && h_verify)) {
    printf("ERROR: Couldn't create host buffers\n");
    return false;
  }

  test_fft(iterations, false); // FFTを実行. 第二引数をtrueにすると, 逆FFTを実行


  // Free the resources allocated
  cleanup();

  return 0;
}

// FFTのテストデータを生成し, カーネルとデータをやり取りする部分
void test_fft(int iterations, bool inverse) {
  printf("Launching");
  if (inverse) 
	printf(" inverse");
  printf(" FFT transform for %d iterations\n", iterations);

  float dt = 1.0/(10*1000*2.56);//サンプリング間隔,(サンプリング周波数2.56x10KHz=25.6KHz)
  float rad = 2*M_PI;    //2π
  float t = 0.0;

  // 合成波をセットする
  // デフォルトでは1600Hzと8000Hz
  for (int i = 0; i < iterations; i++) {
    t = 0.0;
    for (int j = 0; j < N; j++) {
      t = t + dt;
      // 実数部
      h_verify[coord(i, j)].x = h_inData[coord(i, j)].x 
        = (sin(rad*1600*t) + sin(rad*8000*t));
      // 虚数部は0.0にしておく
      h_verify[coord(i, j)].y = h_inData[coord(i, j)].y 
        = 0.0;
    }
  }

  // Create device buffers - assign the buffers in different banks for more efficient
  // memory access 
 
  d_inData = clCreateBuffer(context, CL_MEM_READ_WRITE, sizeof(float2) * N * iterations, NULL, &status);
  checkError(status, "Failed to allocate input device buffer\n");
  d_outData = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_BANK_2_ALTERA, sizeof(float2) * N * iterations, NULL, &status);
  checkError(status, "Failed to allocate output device buffer\n");

  // Copy data from host to device

  status = clEnqueueWriteBuffer(queue, d_inData, CL_TRUE, 0, sizeof(float2) * N * iterations, h_inData, 0, NULL, NULL);
  checkError(status, "Failed to copy data to device");

  // Can't pass bool to device, so convert it to int
  int inverse_int = inverse;

  // Set the kernel arguments
  status = clSetKernelArg(kernel, 0, sizeof(cl_mem), (void *)&d_inData);
  checkError(status, "Failed to set kernel arg 0");
  status = clSetKernelArg(kernel, 1, sizeof(cl_mem), (void *)&d_outData);
  checkError(status, "Failed to set kernel arg 1");
  status = clSetKernelArg(kernel, 2, sizeof(cl_int), (void*)&iterations);
  checkError(status, "Failed to set kernel arg 2");
  status = clSetKernelArg(kernel, 3, sizeof(cl_int), (void*)&inverse_int);
  checkError(status, "Failed to set kernel arg 3");

  printf(inverse ? "\tInverse FFT" : "\tFFT");
  printf(" kernel initialization is complete.\n");

  // Get the iterationstamp to evaluate performance
  double time = getCurrentTimestamp();

  // Launch the kernel - we launch a single work item hence enqueue a task
  status = clEnqueueTask(queue, kernel, 0, NULL, NULL);
  checkError(status, "Failed to launch kernel");

  // Record execution time
  time = getCurrentTimestamp() - time;

  // Wait for command queue to complete pending events
  status = clFinish(queue);
  checkError(status, "Failed to finish");

  // Copy results from device to host
  status = clEnqueueReadBuffer(queue, d_outData, CL_TRUE, 0, sizeof(float2) * N * iterations, h_outData, 0, NULL, NULL);
  checkError(status, "Failed to copy data from device");

  printf("\tProcessing time (FPGA) = %.4fms\n", (float)(time * 1E3));
  double gpoints_per_sec = ((double) iterations * N / time) * 1E-9;
  double gflops = 5 * N * (log((float)N)/log((float)2))/(time / iterations * 1E9);
  printf("\tThroughput = %.4f Gpoints / sec (%.4f Gflops)\n", gpoints_per_sec, gflops);

  // ビット反転
  // FFTはビット逆順に出力するので, データを並び替える.
  //
  // 琉球大和田先生の解説が詳しい.
  // http://www.ie.u-ryukyu.ac.jp/~wada/design07/spec_e.html
  float2 *temp = (float2 *)alloca(sizeof(float2) * N);
  for (int i = 0; i < N; i++) temp[i] = h_outData[i];
  for (int i = 0; i < N; i++) {
    int fwd = i;
    int bit_rev = 0;
    for (int j = 0; j < LOGN; j++) {
      bit_rev <<= 1;
      bit_rev |= fwd & 1;
      fwd >>= 1;
    }
    h_outData[i] = temp[bit_rev];
  }

  // FFTのスペクトルを.csv形式で出力
  FILE *fp;

  fp = fopen( "output.csv", "w");

  printf("      [Hz]      [Pow]\n");
  for (int j = 0; j < N/2; j++) {
    double magnitude = (double)h_outData[coord(0, j)].x * (double)h_outData[coord(0, j)].x +  
                              (double)h_outData[coord(0, j)].y * (double)h_outData[coord(0, j)].y;
    printf("%6.7f %4.7f\n", j*(1.0/dt)/N, magnitude);
    fprintf( fp, "%6.7f %4.7f\n", j*(1.0/dt)/N, magnitude);
  }

  fclose( fp);
}


/////// HELPER FUNCTIONS ///////

// provides a linear offset in the input array
int coord(int iteration, int i) {
  return iteration * N + i;
}


// FFTをソフトウェアで実行するルーチン（今回は未使用, 参考までに) ----------------------------------
// fourier_stage()を再帰的に呼び出す
//
void fourier_transform_gold(bool inverse, const int lognr_points, double2 *data) {
   const int nr_points = 1 << lognr_points;

   // The inverse requires swapping the real and imaginary component
   
   if (inverse) {
      for (int i = 0; i < nr_points; i++) {
         double tmp = data[i].x;
         data[i].x = data[i].y;
         data[i].y = tmp;;
      }
   }
   // Do a FT recursively
   fourier_stage(lognr_points, data);

   // The inverse requires swapping the real and imaginary component
   if (inverse) {
      for (int i = 0; i < nr_points; i++) {
         double tmp = data[i].x;
         data[i].x = data[i].y;
         data[i].y = tmp;;
      }
   }

   // Do the bit reversal

   double2 *temp = (double2 *)alloca(sizeof(double2) * nr_points);
   for (int i = 0; i < nr_points; i++) temp[i] = data[i];
   for (int i = 0; i < nr_points; i++) {
      int fwd = i;
      int bit_rev = 0;
      for (int j = 0; j < lognr_points; j++) {
         bit_rev <<= 1;
         bit_rev |= fwd & 1;
         fwd >>= 1;
      }
      data[i] = temp[bit_rev];
   }
}

void fourier_stage(int lognr_points, double2 *data) {
   int nr_points = 1 << lognr_points;
   if (nr_points == 1) return;
   double2 *half1 = (double2 *)alloca(sizeof(double2) * nr_points / 2);
   double2 *half2 = (double2 *)alloca(sizeof(double2) * nr_points / 2);
   for (int i = 0; i < nr_points / 2; i++) {
      half1[i] = data[2 * i];
      half2[i] = data[2 * i + 1];
   }
   fourier_stage(lognr_points - 1, half1);
   fourier_stage(lognr_points - 1, half2);
   for (int i = 0; i < nr_points / 2; i++) {
      data[i].x = half1[i].x + cos (2 * M_PI * i / nr_points) * half2[i].x + sin (2 * M_PI * i / nr_points) * half2[i].y;
      data[i].y = half1[i].y - sin (2 * M_PI * i / nr_points) * half2[i].x + cos (2 * M_PI * i / nr_points) * half2[i].y;
      data[i + nr_points / 2].x = half1[i].x - cos (2 * M_PI * i / nr_points) * half2[i].x - sin (2 * M_PI * i / nr_points) * half2[i].y;
      data[i + nr_points / 2].y = half1[i].y + sin (2 * M_PI * i / nr_points) * half2[i].x - cos (2 * M_PI * i / nr_points) * half2[i].y;
   }
}
// --------------------------------------------------------------------------------

bool init() {
  cl_int status;

  if(!setCwdToExeDir()) {
    return false;
  }

  // Get the OpenCL platform.
  platform = findPlatform("Altera");
  if(platform == NULL) {
    printf("ERROR: Unable to find Altera OpenCL platform\n");
    return false;
  }

  // Query the available OpenCL devices.
  scoped_array<cl_device_id> devices;
  cl_uint num_devices;

  devices.reset(getDevices(platform, CL_DEVICE_TYPE_ALL, &num_devices));

  // We'll just use the first device.
  device = devices[0];

  // Create the context.
  context = clCreateContext(NULL, 1, &device, &oclContextCallback, NULL, &status);
  checkError(status, "Failed to create context");

  // Create the command queue.
  queue = clCreateCommandQueue(context, device, CL_QUEUE_PROFILING_ENABLE, &status);
  checkError(status, "Failed to create command queue");

  // Create the program.
  std::string binary_file = getBoardBinaryFile("fft1d", device);
  printf("Using AOCX: %s\n\n", binary_file.c_str());
  program = createProgramFromBinary(context, binary_file.c_str(), &device, 1);

  // Build the program that was just created.
  status = clBuildProgram(program, 0, NULL, "", NULL, NULL);
  checkError(status, "Failed to build program");

  // Create the kernel - name passed in here must match kernel name in the
  // original CL file, that was compiled into an AOCX file using the AOC tool
  const char *kernel_name = "fft1d";  // Kernel name, as defined in the CL file
  kernel = clCreateKernel(program, kernel_name, &status);
  checkError(status, "Failed to create kernel");

  return true;
}

// Free the resources allocated during initialization
void cleanup() {
  if(kernel) 
    clReleaseKernel(kernel);  
  if(program) 
    clReleaseProgram(program);
  if(queue) 
    clReleaseCommandQueue(queue);
  if(context) 
    clReleaseContext(context);
  if(h_inData)
	alignedFree(h_inData);
  if (h_outData)
	alignedFree(h_outData);
  if (h_verify)
	alignedFree(h_verify);
  if (d_inData)
	clReleaseMemObject(d_inData);
  if (d_outData) 
	clReleaseMemObject(d_outData);
}



